/*
 * Memory.c
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 *
 * A generic memory manager that can be used to allocate memory
 * from larger memory blocks allocated from the heap or shared
 * memory. This is a fairly simple technique and is patterned
 * after K&R2 page 185.
 */

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include <string.h>

#include "Memory.h"

struct memory {
	unsigned long size;	/* size of chunk, includes this struct. */
	struct memory *next;
};

#define MEMORY_SIZEOF			(sizeof (unsigned long))
#define MEMORY_MINIMUM_CHUNK_SIZE	(sizeof (struct memory) - MEMORY_SIZEOF)

int
MemoryInit(void *block, unsigned long size)
{
	struct memory *chunk;
	struct memory *head = (struct memory *) block;

	if (block == (void *) 0 || size < 2 * sizeof *chunk)
		return -1;

	head->size = size;
	head->next = chunk = &head[1];

	chunk->size = size - sizeof *head;
	chunk->next = (struct memory *) 0;

	return 0;
}

unsigned long
MemorySizeOf(void *chunk)
{
	if (chunk == (void *) 0)
		return 0;

	return ((unsigned long *) chunk)[-1] - MEMORY_SIZEOF;
}

void *
MemoryAllocate(void *block, unsigned long size)
{
	unsigned long excess;
	struct memory *best, *upper;
	struct memory *prev = (struct memory *) block;

	if (block == (void *) 0)
		return (void *) 0;

	/* Insure the requested size includes accounting data. */
	if (size < MEMORY_MINIMUM_CHUNK_SIZE)
		size = MEMORY_MINIMUM_CHUNK_SIZE;
	size += MEMORY_SIZEOF;

	/* Align memory to size of long units. */
	size = (((size-1) / sizeof (long)) + 1) * sizeof (long);

	/* Walk list looking for smallest chunk for the request.
	 * "best" points to one entry before the best chunk.
	 */
	for (best = prev; prev->next != (struct memory *) 0; prev = prev->next) {
		unsigned long chunksize = prev->next->size;
		if (size <= chunksize && chunksize < best->size)
			best = prev;
	}

	/* Empty list or is the requested size too large? */
	if (best->next == (struct memory *) 0 || best->next->size < size)
		return (void *) 0;

	/* "best" points to one entry before the best chunk. */
	prev = best;
	best = prev->next;

	/* "prev" now points to one entry before the best chunk
	 * and "best" points to the best chunk.
	 */

	/* Can the best chunk be reduced in size? */
	excess = best->size - size;

	if (sizeof *best <= excess) {
		/* Split best chunk into unequal halves. */
		best->size = size;

		/* Point to upper half. */
		upper = (struct memory *) &((char *) best)[size];

		/* Insert upper half chunk into free list. */
		upper->size = excess;
		upper->next = best->next;
		best->next = upper;
	}

	/* Remove best chunk from free list. */
	prev->next = best->next;

	return (void *) ((char *) best + MEMORY_SIZEOF);
}

void
MemorySet(void *chunk, int byte)
{
	memset(chunk, byte, MemorySizeOf(chunk));
}

void
MemoryFree(void *block, void *allocated)
{
	struct memory *prev;
	struct memory *here = (struct memory *) block;
	struct memory *chunk = (struct memory *) ((char *) allocated - MEMORY_SIZEOF);

	if (block == (void *) 0 || allocated == (void *) 0)
		return;

	/* Allocated chunk NOT from this block. */
	if (allocated <= block && (char *) here + here->size <= (char *) allocated)
		return;

	/* Find chunk immediately after the allocated one in memory. */
	for ( ; here != (struct memory *) 0; prev = here, here = here->next) {
		if (chunk < here) {
			/* Insert chunk before here. */
			chunk->next = here;
			prev->next = chunk;
			break;
		}
	}

	/* Add this chunk to the end of the free list? */
	if (here < chunk) {
		chunk->next = here;
		prev->next = chunk;
	}

	/* Is the next chunk adjacent to this one? */
	if ((char *) chunk + chunk->size == (char *) chunk->next) {
		/* Merge this chunk and the next. */
		chunk->size += chunk->next->size;
		chunk->next = chunk->next->next;
	}

	/* Is the previous chunk adjacent to this one? */
	if ((char *) prev + prev->size == (char *) chunk) {
		/* Merge previous chunk and this one. */
		prev->size += chunk->size;
		prev->next = chunk->next;
	}
}

void *
MemoryReallocate(void *block, void *allocated, unsigned long size)
{
	void *replace;

	if (block == (void *) 0)
		return (void *) 0;

	if (allocated == (void *) 0)
		return MemoryAllocate(block, size);

	if (size <= MemorySizeOf(allocated))
		return allocated;

	if ((replace = MemoryAllocate(block, size)) == (void *) 0)
		return (void *) 0;

	memcpy(replace, allocated, MemorySizeOf(allocated));
	MemoryFree(block, allocated);

	return replace;
}

/*
 * Similar to MemoryReallocate(), but does NOT preserve the
 * contents of the previous allocated memory block.
 */
void *
MemoryResize(void *block, void *allocated, unsigned long size)
{
	if (block == (void *) 0)
		return (void *) 0;

	if (allocated == (void *) 0)
		return MemoryAllocate(block, size);

	if (size <= MemorySizeOf(allocated))
		return allocated;

	MemoryFree(block, allocated);

	return MemoryAllocate(block, size);
}

unsigned long
MemoryAvailable(void *block)
{
	unsigned long space;
	struct memory *here = (struct memory *) block;

	if (block == (void *) 0)
		return 0;

	for (space = 0, here = here->next; here != (struct memory *) 0; here = here->next) {
		space += here->size - MEMORY_SIZEOF;
	}

	return space;
}

#if defined(TEST)

#include <stdio.h>
#include <stdlib.h>

#define MARKER		__FILE__,__LINE__

void
notNull(void *ptr, char *file, long line)
{
	if (ptr == (void *) 0) {
		printf("Expected non-null pointer at %s:%ld\n", file, line);
		exit(1);
	}
}

void
isNull(void *ptr, char *file, long line)
{
	if (ptr != (void *) 0) {
		printf("Expected null pointer at %s:%ld\n", file, line);
		exit(1);
	}
}

void
expectedChunkSize(void *block, void *chunk, unsigned long size, char *file, long line)
{
	unsigned long chunksize = ((unsigned long *) chunk)[-1];
	printf("Memory space available: %lu\n", MemoryAvailable(block));
	if (chunksize != size) {
		printf("Unexpected chunk size %d at %s:%ld\n", chunksize, file, line);
		exit(1);
	}
}

int
main(int argc, char **argv)
{
	void *block, *first, *a, *b, *c;

	setvbuf(stdout, (char *) 0, _IOLBF, BUFSIZ);

	block = malloc(100);
	notNull(block, MARKER);
	(void) MemoryInit(block, 100);

	/* TEST 1
	 * show free space
	 */
	printf("Memory space available: %lu\n", MemoryAvailable(block));

	/* TEST 2
	 * allocate & free should result in initial free space;
	 */
	first = MemoryAllocate(block, 20);
	notNull(first, MARKER);

	expectedChunkSize(block, first, 20 +  MEMORY_SIZEOF, MARKER);
	MemoryFree(block, first);
	expectedChunkSize(block, first, 100 - sizeof (struct memory), MARKER);

	/* TEST 3
	 * allocate & free in different order
	 */
	first = MemoryAllocate(block, 0);
	notNull(first, MARKER);
	expectedChunkSize(block, first, sizeof (struct memory), MARKER);

	a = MemoryAllocate(block, 20);
	notNull(a, MARKER);
	expectedChunkSize(block, a, 20 +  MEMORY_SIZEOF, MARKER);

	b = MemoryAllocate(block, 20);
	notNull(b, MARKER);
	expectedChunkSize(block, b, 20 +  MEMORY_SIZEOF, MARKER);

	c = MemoryAllocate(block, 20);
	notNull(c, MARKER);
	expectedChunkSize(block, c, 20 +  MEMORY_SIZEOF, MARKER);

	MemoryFree(block, c);
	MemoryFree(block, a);
	MemoryFree(block, b);
	expectedChunkSize(block, a, 100 - 2 * sizeof (struct memory), MARKER);

	MemoryFree(block, first);
	expectedChunkSize(block, first, 100 - sizeof (struct memory), MARKER);

	/* TEST 4
	 * allocate beyond free space should result in null
	 */
	first = MemoryAllocate(block, 0);
	notNull(first, MARKER);
	expectedChunkSize(block, first, sizeof (struct memory), MARKER);

	a = MemoryAllocate(block, 20);
	notNull(a, MARKER);
	expectedChunkSize(block, a, 20 +  MEMORY_SIZEOF, MARKER);

	b = MemoryAllocate(block, 40);
	notNull(b, MARKER);
	expectedChunkSize(block, b, 40 +  MEMORY_SIZEOF, MARKER);

	c = MemoryAllocate(block, 60);
	isNull(c, MARKER);

	MemoryFree(block, a);
	MemoryFree(block, b);
	expectedChunkSize(block, a, 100 - 2 * sizeof (struct memory), MARKER);

	MemoryFree(block, first);
	expectedChunkSize(block, first, 100 - sizeof (struct memory), MARKER);

	free(block);

	printf("OK\n");

	return 0;
}

#endif
