/*
 * SharedHash.c
 *
 * Shared memory hash table with long term disk storage.
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 */

/*
 * Maximum number of linear probes when searching the hash table
 * before swapping the oldest entry out to disk. I recommend a
 * small number betweeb 3 and 20. I think it is better to swap
 * out to disk an entry than spend too much time searching.
 *
 * This could be a module configuration option, but I am of the
 * opinion too many choices can result in poor performance. Best
 * to keep this as a small and reasonable constant.
 */
#ifndef MAX_PROBES
#define MAX_PROBES		8
#endif

/*
 * The average length of a key string. Used in computing extra
 * space required to store key strings in shared memory. When
 * key string space is exhausted, some hash entries are flushed
 * to disk to make room for newer entries.
 */
#ifndef AVERAGE_KEY_LENGTH
#define AVERAGE_KEY_LENGTH	20
#endif

/*
 * The longest supported key string. This should account for IP
 * addresses in ASCII, user names, virtual host names like
 * www.domain.tld and subdomain.domain.tld.  From what I recall,
 * domain names can be 22 characters plus a dot plus a 4 character
 * top-level-domain. I've allowed for extra space to account for
 * machine or sub-domain names.
 */
#ifndef MAX_KEY_LENGTH
#define MAX_KEY_LENGTH		128
#endif

#define ENABLE_LOG		1

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#include "httpd.h"
#include "http_log.h"
#include "http_conf_globals.h"
extern server_rec *watchMainServer;

#include "Mutex.h"
#include "Shared.h"
#include "Memory.h"
#include "SharedHash.h"

const char shLockFile[] = "SharedHash.lock";

int
shLock(struct shTable *table)
{
	if (table == (struct shTable *) 0)
		return -1;

	return MutexLock(table->mutex);
}

int
shUnlock(struct shTable *table)
{
	if (table == (struct shTable *) 0)
		return -1;

	return MutexUnlock(table->mutex);
}

/*
 * Store a shared memory hash table entry to disk.
 */
static void
shStore(struct shTable *table, struct shEntry *entry)
{
	FILE *fp;

	if (entry == (struct shEntry *) 0 || entry->key == (char *) 0)
		return;

	/* ASSUME the workdir string has a trailing directory separator. */
	strcpy(table->pathname, table->workdir);
	strncat(table->pathname, entry->key, MAX_KEY_LENGTH);

	if ((fp = fopen(table->pathname, "w")) != (FILE *) 0) {
		fprintf(
			fp, "%lu %lu %lu %lu\n",
			entry->ifInOctets,
			entry->ifOutOctets,
			entry->ifRequests,
			entry->ifDocuments
		);
		fclose(fp);
	}
}

/*
 * Flush the entier shared memeory hash table to disk, releasing
 * the memory assigned for all the key strings.
 */
static void
shFlushAll(struct shTable *table)
{
	struct shEntry *here, *stop;

	stop = table->array + table->size;

	for (here = table->array; here < stop; ++here) {
		shStore(table, here);
		MemoryFree(table->shared, here->key);
		here->key = (char *) 0;
	}
}

/*
 * Flush to disk a consective series of shared memory hash table
 * entries either side of the given index.
 */
static void
shFlush(struct shTable *table, int index)
{
	int count = 0;
	struct shEntry *here, *stop;

	table->info->flushes++;
	stop = table->array + table->size;

	/* Flush forward until next null. */
	here = table->array + (index + 1) % table->size;

	while (here->key != (char *) 0) {
		shStore(table, here);
		MemoryFree(table->shared, here->key);
		here->key = (char *) 0;
		count++;

		if (stop <= ++here)
			here = table->array;
	}

	/* Flush backward until previous null. */
	here = table->array + index;

	while (here->key != (char *) 0) {
		shStore(table, here);
		MemoryFree(table->shared, here->key);
		here->key = (char *) 0;
		count++;

		if (--here < table->array)
			here = stop - 1;
	}

	if (count <= 0)
		shFlushAll(table);
}

/*
 * Fetch a shared memory hash table entry from disk. If we run out of
 * shared memory space, then flush some entries to disk before fetching
 * the entry for the key.
 */
static struct shEntry *
shFetch(struct shTable *table, const char *key, int index)
{
	FILE *fp;
	int keylen = (int) strlen(key) + 1;
	struct shEntry *entry = &table->array[index];

	entry->key = MemoryResize(table->shared, entry->key, keylen);
	if (entry->key == (char *) 0) {
		shFlush(table, index);
		entry->key = MemoryAllocate(table->shared, keylen);
		if (entry->key == (char *) 0) {
			return (struct shEntry *) 0;
		}
	}

	/* ASSUME the workdir string has a trailing directory separator. */
	strcpy(table->pathname, table->workdir);
	strncat(table->pathname, key, MAX_KEY_LENGTH);
	strcpy(entry->key, key);

	if ((fp = fopen(table->pathname, "r")) != (FILE *) 0) {
		(void) fscanf(
			fp, "%lu %lu %lu %lu",
			&entry->ifInOctets,
			&entry->ifOutOctets,
			&entry->ifRequests,
			&entry->ifDocuments
		);
		fclose(fp);
	}

	return entry;
}

unsigned long
shHashCode(unsigned long hash, const char *k)
{
	for ( ; *k != '\0'; ++k)
		hash = (hash * 37) + *k;

	return hash;
}

int
shContainsKey(struct shTable *table, const char *key)
{
	int i;
	struct shEntry *array;
	unsigned long hash, start;

	if (table == (struct shTable *) 0)
		return 0;

	array = table->array;
	start = hash = shHashCode(0, key) % table->size;

	for (i = 0; i <= MAX_PROBES; ++i) {
		if (array[hash].key == (char *) 0)
			return 0;
		if (strcmp(key, array[hash].key) == 0)
			return 1;

		/* Linear probe through the table. */
		hash = (hash + 1) % table->size;
	}

	return 0;
}

/*
 * Return a pointer to the shared memory hash table entry for the key.
 */
struct shEntry *
shGetLockedEntry(struct shTable *table, const char *key)
{
	int i;
	const char *k;
	struct shEntry *entry;
	unsigned long hash, start;

	/* A simple hash function. Factors of either 31 or 37 are good
	 * values for text strings. If the table size is prime, then
	 * we can double-hash or linear probe through entire table.
	 */
	start = hash = shHashCode(0, key) % table->size;

	if (MutexLock(table->mutex))
		return (struct shEntry *) 0;

	/* Perform a limited number of linear probes; no point in
	 * looking too far, better just to swap an entry out to disk.
	 */
	for (i = 0; i <= MAX_PROBES; ++i) {
		/* Free slot in the hash table? */
		if (table->array[hash].key == (char *) 0) {
			if ((entry = shFetch(table, key, hash)) == (struct shEntry *) 0) {
#ifdef ENABLE_LOG
				ap_log_error(
					APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
					"shFetch(%lx, key=\"%s\", hash=%d) failed filling empty entry",
					(unsigned long) table, key, hash
				);
#endif
				MutexUnlock(table->mutex);
			} else {
				table->info->found++;
			}

			return entry;
		}

		/* Found existing key? */
		if (strcmp(key, table->array[hash].key) == 0) {
			table->info->found++;
			return &table->array[hash];
		}

		/* Linear probe through the table. */
		hash = (hash + 1) % table->size;
		table->info->probes++;
	}

	/* No matching key found within probing limit. Replace the
	 * initial hash starting point.
	 */
	shStore(table, &table->array[start]);
	if ((entry = shFetch(table, key, start)) == (struct shEntry *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_INFO, watchMainServer,
			"shFetch(%lx, key=\"%s\", start=%d) failed replacing entry",
			(unsigned long) table, key, start
		);
#endif
		MutexUnlock(table->mutex);
	} else {
		table->info->faults++;
	}

	return entry;
}

/*
 * Create a shared memory table.
 */
struct shTable *
shCreate(struct pool *p, int size, const char *workdir)
{
	size_t bytes;
	const char *lockfile;
	struct shTable *table;

	if (size <= MAX_PROBES) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_NOERRNO|APLOG_ERR, watchMainServer,
			"shCreate() hash table size too small (%d)",
			size
		);
#endif
		goto error0;
	}

	lockfile = ap_pstrcat(p, workdir, shLockFile, (char *) 0);
	if (lockfile == (const char *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"shCreate() failed to allocate lockfile string"
		);
#endif
		goto error0;
	}

	/* Allocate enough storage for the table management elements,
	 * which includes a pathname construction buffer.
	 */
	table = ap_pcalloc(p, sizeof *table + strlen(workdir) + 1 + MAX_KEY_LENGTH + 1);
	if (table == (struct shTable *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"shCreate() failed to allocate shTable structure"
		);
#endif
		goto error0;
	}

	/* Space for global data. */
	bytes = sizeof *table->info;

	/* We need enough room for an array of N entries, with an
	 * average key length of M bytes.
	 */
	bytes += size * (sizeof *table->array + AVERAGE_KEY_LENGTH);

	if ((table->shared = SharedCreate(bytes)) == (void *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"SharedCreate(%lu) failed", bytes
		);
#endif
		goto error0;
	}

	if (SharedPermission(table->shared, 0600, ap_user_id, ap_group_id) < 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"SharedPermission(%lx, 0600, %d=%s, %d=%s) failed",
			(unsigned long) table->shared,
			ap_user_id, watchGetUidName(p, ap_user_id),
			ap_group_id, watchGetGidName(p, ap_group_id)
		);
#endif
		goto error1;
	}

	if (MemoryInit(table->shared, bytes) != 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MemoryInit(%lx, %lu) failed",
			(unsigned long) table->shared, bytes
		);
#endif
		goto error1;
	}

	if ((table->mutex = MutexCreate(lockfile)) == (void *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MutexCreate(%s) failed", lockfile
		);
#endif
		goto error1;
	}

	if (MutexPermission(table->mutex, 0600, ap_user_id, ap_group_id) < 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MutexPermission(%lx, 0600, %d=%s, %d=%s) failed",
			(unsigned long) table->mutex,
			ap_user_id, watchGetUidName(p, ap_user_id),
			ap_group_id, watchGetGidName(p, ap_group_id)
		);
#endif
		goto error2;
	}

	table->info = (struct shInfo *) MemoryAllocate(table->shared, sizeof *table->info);
	if (table->info == (struct shInfo *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MemoryAllocate(%lx, %lu) #1 failed",
			(unsigned long) table->shared, sizeof *table->info
		);
#endif
		goto error2;
	}

	table->array = (struct shEntry *) MemoryAllocate(table->shared, size * sizeof *table->array);
	if (table->array == (struct shEntry *) 0) {
#ifdef ENABLE_LOG
		ap_log_error(
			APLOG_MARK, APLOG_ERR, watchMainServer,
			"MemoryAllocate(%lx, %lu) #2 failed",
			(unsigned long) table->shared, size * sizeof *table->array
		);
#endif
		goto error2;
	}

	MemorySet(table->info, 0);
	MemorySet(table->array, 0);

	table->pathname = (char *) table + sizeof *table;
	table->workdir = workdir;
	table->size = size;

	return table;
error2:
	MutexDestroy(table->mutex);
error1:
	SharedDestroy(table->shared);
error0:
	return (struct shTable *) 0;
}

/*
 * Save the entrie shared memory array to disk.
 */
void
shDestroy(void *data)
{
	struct shTable *table = (struct shTable *) data;

	if (table != (struct shTable *) 0) {
		shFlushAll(table);
		MutexDestroy(table->mutex);
		SharedDestroy(table->shared);
	}
}

