/*
 * mod_watch.c for Apache 1.3
 *
 * Copyright 2001, 2002 by Anthony Howe.  All rights reserved.
 *
 *
 * LICENSE
 * -------
 *
 * 1) Non-commercial redistribution and use of the original source
 * distribution, without modification, is permitted provided you retain
 * this notice, the author's copyright notice above, and the disclaimer
 * below.
 *
 * 2) Commercial redistribution and use of the original source
 * distribution and/or derived binaries, with or without modification,
 * is permitted with a) the prior written permission from the author,
 * b) financial remuneration to the author, c) the above copyright
 * notice is retained and reproduced by the binary, and d) the
 * disclaimer below appears in the documentation and/or other materials
 * provided with the distribution.
 *
 *
 * DISCLAIMER
 * ----------
 *
 * THIS SOFTWARE IS PROVIDE "AS IS" AND ANY EXPRESSED OR IMPLIED WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO WAY SHALL THE
 * AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#ifndef SPOOLDIR
#define SPOOLDIR		"/var/spool/mod_watch/"
#endif

/*
 * The size of the hash table should be a small prime number:
 *
 *	449, 673, 991, 1409, 2411
 *
 * Using a prime number for the table size means that double
 * hashing or linear-probing can visit all possible entries.
 *
 * This is NOT a runtime option, because its not something I
 * want people to play with unless absolutely necessary.
 */
#ifndef TABLE_SIZE
#define TABLE_SIZE		991
#endif

/***********************************************************************
 *** No configuration below this point.
 ***********************************************************************/

#define MODULE			"mod_watch"
#define AUTHOR			"achowe@snert.com"
#define VERSION			"3.6"
#define COPYRIGHT_DATE		"2001, 2002"
#define NDEBUG			1

#include "httpd.h"
#include "http_core.h"
#include "http_config.h"
#include "http_conf_globals.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "ap_config.h"

#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <dirent.h>
#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netdb.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#define UNSET			-1
#define MERGE(p, c)		(c == UNSET ? p : c)
#define MERGE_PTR(p, c)		((void *) c == (void *) 0 ? p : c)

module MODULE_VAR_EXPORT watch_module;

/*
 *	counter32	ifInOctets	bytes received from POST requests
 *	counter32	ifOutOctets	bytes sent in response to requests
 *	counter32	ifRequests	request / hit counter
 *	counter32	ifDocument	document hit counter
 *	counter32	ifRefused	(future) request refused by policy
 *	integer		delay		(future) current response delay/credit
 *	integer		concurrent	(future) number of concurrent requests
 */
unsigned long ifInOctets, ifOutOctets, ifRequests, ifDocuments;
unsigned long reqInOctets, reqOutOctets, reqRequests, reqDocuments;

#define WATCH_SERVER		1
#define WATCH_VHOST		2
#define WATCH_OWNER		4
#define WATCH_REMOTE_IP		8
#define WATCH_LOCATION		16
#define WATCH_DEFAULT		32768
#define WATCH_ALL		(~0)

#define WATCH_INFO		1
#define WATCH_TABLE		2
#define WATCH_COPYRIGHT		4

server_rec *watchMainServer;

static int isWatchHandler = 0;
static int watch_what = WATCH_DEFAULT | WATCH_VHOST | WATCH_SERVER;
static int watch_not_logged = 0;

static char *spoolDirectoryRun;
static char *spoolDirectoryConf;

static const char SERVER[] = "SERVER";
static const char LOCKFILE[] = "mod_watch.lock";

static array_header *watchSpoolFileList;
static array_header *watchSpoolFileConf;

struct watchConfDir {
	int index;
};

static array_header *document_conf;
static array_header *document_list;
static const char *documents_def[] = {
	"text/html",
	"text/plain",
	"text/xml",
	"application/pdf",
	"text/enriched",
	"text/richtext",
	"application/*word",
	"application/*excel",
	"application/postscript",
	"application/rtf",
/*	"httpd/unix-directory", */
	(char *) 0
};

typedef struct store_policy {
	const char *name;
	int (*content) (request_rec *);
	void (*update) (request_rec *, const char *);
} store_policy;

static void watchUpdateFiles(request_rec *r, const char *key);
static void watchUpdateHash(request_rec *r, const char *key);
static void (*watchUpdate)(request_rec *, const char *);

static int watchContentFiles(request_rec *r);
static int watchContentHash(request_rec *r);

static store_policy policy_table[] = {
	{ "files", watchContentFiles, watchUpdateFiles },
	{ "hash", watchContentHash, watchUpdateHash },
	{ 0, 0 }
};

/***********************************************************************
 *** Shared Memory Hash Table Routines
 ***********************************************************************/

#include "SharedHash.h"

static struct shTable *shtable;
static int shTableSize = TABLE_SIZE;

/***********************************************************************
 *** Semaphore / Mutex Routines
 ***********************************************************************/

#include "Mutex.h"

static void *mutex;

/***********************************************************************
 *** Network Table Routines
 ***********************************************************************/

#include "NetworkTable.h"

static array_header *networkIncludeConf;
static array_header *networkIncludeList;
static array_header *networkExcludeConf;
static array_header *networkExcludeList;

/***********************************************************************
 *** Support Routines
 ***********************************************************************/

#define UNKNOWN_UID		((uid_t) ~0)
#define UNKNOWN_GID		((gid_t) ~0)

/*
 */
const char *
watchGetUidName(struct pool *p, uid_t uid)
{
	if (uid == UNKNOWN_UID)
		return "unknown-user";

#if defined(WIN32) || defined(NETWARE)
	return "default-user";
#else
{
	struct passwd *pw;

	if ((pw = getpwuid(uid)) != (struct passwd *) 0)
		return ap_pstrdup(p, pw->pw_name);

	return "unknown-user";
}
#endif
}

const char *
watchGetGidName(struct pool *p, gid_t gid)
{
	if (gid == UNKNOWN_GID)
		return "unknown-group";

#if defined(WIN32) || defined(NETWARE)
	return "default-group";
#else
{
	struct group *gr;

	if ((gr = getgrgid(gid)) != (struct group *) 0)
		return ap_pstrdup(p, gr->gr_name);

	return "unknown-group";
}
#endif
}

static const char *
setDirectory(struct pool *p, char *arg, char **directory)
{
	struct stat buf;

        arg = ap_server_root_relative(p, arg);
	if (stat(arg, &buf) != 0 || !S_ISDIR(buf.st_mode))
		return "Pathname must be a directory.";

	/* Append trailing slash. */
	if (strrchr(arg, '/')[1] != '\0')
		*directory = ap_pstrcat(p, arg, "/", (char *) 0);
	else
		*directory = ap_pstrdup(p, arg);

	return (const char *) 0;
}

static const char *
setUnsignedInt(const char *arg, unsigned int *number)
{
	char *stop;
	long value;

	if (arg == (char *) 0 || *arg == '\0')
		return "Integer unspecified";

	value = strtol(arg, &stop, 10);

	if (*stop != '\0')
		return "Not a decimal integer";

	if (value < 0)
		return "Not a positive integer";

	if (UINT_MAX < value)
		return "Integer too large";

	*number = (unsigned int) value;

	return (const char *) 0;
}

/***********************************************************************
 ***  Create, Init, and Merge Routines
 ***********************************************************************/

static void *
watchCreateDir(struct pool *p, char *dir)
{
	struct watchConfDir *conf = ap_pcalloc(p, sizeof *conf);

	conf->index = UNSET;

	return (void *) conf;
}

/*
 *
 */
static void
watchInit(server_rec *s, struct pool *p)
{
	/* Handy to have for error logging. */
	watchMainServer = s;

        ap_add_version_component(MODULE"/"VERSION);

	/* If no WatchDocuments specified, use defaults. */
	if (document_conf == (array_header *) 0) {
		const char **list, **mime;

		document_conf = ap_make_array(
			p,
			sizeof documents_def / sizeof *documents_def,
			sizeof *documents_def
		);

		for (list = documents_def; *list != (char *) 0; ++list) {
			mime = (const char **) ap_push_array(document_conf);
			if (mime == (const char **) 0)
				exit(APEXIT_INIT);
			*mime = *list;
		}
	}

	if (spoolDirectoryConf == (char *) 0) {
		const char *msg = setDirectory(p, SPOOLDIR, &spoolDirectoryConf);
		if (msg != (const char *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_CRIT, watchMainServer,
				"%s: %s", msg, SPOOLDIR
			);
			exit(APEXIT_INIT);
		}
	}

	/* Prepare for next restart. */
        document_list = document_conf;
        document_conf = (array_header *) 0;

	networkIncludeList = networkIncludeConf;
	networkIncludeConf = (array_header *) 0;

	networkExcludeList = networkExcludeConf;
	networkExcludeConf = (array_header *) 0;

	watchSpoolFileList = watchSpoolFileConf;
	watchSpoolFileConf = (array_header *) 0;

	spoolDirectoryRun = spoolDirectoryConf;
	spoolDirectoryConf = (char *) 0;

	if (watchUpdate == (void (*)(request_rec *, const char *)) 0)
		watchUpdate = watchUpdateHash;

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchInit(%lx, %lx) watchUpdate=%lx watchUpdateHash=%lx watchUpdateFiles=%lx",
		(unsigned long) watchUpdate, (unsigned long) watchUpdateHash, (unsigned long) watchUpdateFiles
	);

	if (watchUpdate == watchUpdateFiles) {
		/* Create a mutex used to control access to the weenie files. */
		const char *lockfile = ap_pstrcat(p, spoolDirectoryRun, LOCKFILE, (char *) 0);
		if (lockfile == (const char *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_ERR, watchMainServer,
				"watchInit() failed to allocate lockfile string"
			);
			exit(APEXIT_INIT);
		}

		if ((mutex = MutexCreate(lockfile)) == (void *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_ERR, watchMainServer,
				"MutexCreate(%s) failed in watchInit()", lockfile
			);
			exit(APEXIT_INIT);
		}

		if (MutexPermission(mutex, 0600, ap_user_id, ap_group_id) < 0) {
			ap_log_error(
				APLOG_MARK, APLOG_ERR, watchMainServer,
				"MutexPermission(%lx, 0600, %d=%s, %d=%s) failed in watchInit()",
				(unsigned long) mutex,
				ap_user_id, watchGetUidName(p, ap_user_id),
				ap_group_id, watchGetGidName(p, ap_group_id)
			);
			exit(APEXIT_INIT);
		}
		ap_register_cleanup(p, mutex, MutexDestroy, ap_null_cleanup);
	} else {
		/* Create shared memory hash table. */
		shtable = shCreate(p, shTableSize, spoolDirectoryRun);
		if (shtable == (struct shTable *) 0) {
			ap_log_error(
				APLOG_MARK, APLOG_CRIT, watchMainServer,
				"shCreate(%lx, %lu, %s) failed in watchInit()",
				(unsigned long) p, shTableSize, spoolDirectoryRun
			);
			exit(APEXIT_INIT);
		}
		ap_register_cleanup(p, shtable, shDestroy, ap_null_cleanup);
	}

#ifndef NDEBUG
	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, s,
		"watchMainServer=%s", s->server_hostname
	);
#endif
}

/***********************************************************************
 ***  Handlers
 ***********************************************************************/

/*
 * All of our content handlers return plain text (well almost).
 */
static int
watchFixup(request_rec *r)
{
	isWatchHandler = 0;

	if (r->handler == (char *) 0)
		return DECLINED;

	if (ap_strcasecmp_match(r->handler, "watch-info") == 0) {
		isWatchHandler = WATCH_INFO;
		r->content_type = "text/plain";
	} else if (ap_strcasecmp_match(r->handler, "watch-table") == 0) {
		isWatchHandler = WATCH_TABLE;
		r->content_type = "text/html";
	} else if (ap_strcasecmp_match(r->handler, "watch-copyright") == 0) {
		isWatchHandler = WATCH_COPYRIGHT;
		r->content_type = "text/plain";
	}

	return isWatchHandler == 0 ? DECLINED : OK;
}

static void
watchCounters(request_rec *r)
{
	const char *value;
	unsigned long content_length;

	reqInOctets = 0;
	reqOutOctets = r->bytes_sent;

	reqRequests = 1;
	reqDocuments = 0;

	value = ap_table_get(r->headers_in, "Content-Length");
	if (value != (const char *) 0)
		reqInOctets = strtol(value, (char **) 0, 10);

	/* Update ifDocuments counter. */
	if (r->content_type != (const char *) 0) {
		int i;
		request_rec *rr;
		char **list = (char **) document_list->elts;

		/* Find actual content handler issued. */
		for (rr = r; rr->next != (request_rec *) 0; rr = rr->next)
			;

		for (i = 0; i < document_list->nelts; ++i) {
			if (ap_strcasecmp_match(rr->content_type, list[i]) == 0) {
				reqDocuments = 1;
				break;
			}
		}
	}
}

/*
 * Pick a target, which is either the server name of the request
 * or a specified ~user name or ~SERVER.
 */
static const char *
watchTarget(request_rec *r)
{
	if (r->uri[0] == '/' && r->uri[1] == '~' && ap_isalnum(r->uri[2])) {
		const char *remainder = r->uri + 2;
		return ap_getword(r->pool, &remainder, '/');
	}

	return ap_get_server_name(r);
}

/*
 * Write the content, which is single text line containing the
 * target name, server uptime in seconds, octets in, octets out,
 * requests in, and documents out.
 */
static int
watchContentPrint(request_rec *r, const char *name)
{
	ap_send_http_header(r);

	if (r->header_only)
		return OK;

	ap_rprintf(
		r, "%s %lu %lu %lu %lu %lu\n", name,
		r->request_time - ap_restart_time,
		ifInOctets, ifOutOctets, ifRequests, ifDocuments
	);

	return OK;
}

static int
watchContentCopyright(request_rec *r)
{
	ap_send_http_header(r);

	if (r->header_only)
		return OK;

	ap_rprintf(
		r, MODULE "/" VERSION
		" Copyright " COPYRIGHT_DATE " by Anthony Howe. All rights reserved.\n"
	);

	return OK;
}

static int
watchLog(request_rec *r)
{
	struct watchConfDir *dconf = (struct watchConfDir *) ap_get_module_config(
		r->per_dir_config, &watch_module
	);

	/* Do not count the watch-info reports. */
	if (isWatchHandler) {
		/* The Writing Apache Modules book on p78 states that all the
		 * log phase handlers are executed regardless of whether they
		 * return OK or DECLINED. A log handler can short-circuit the
		 * remaining log handlers if it returns something other than
		 * OK or DECLINED. This is highly dependant on module load
		 * order and could affect the correct opertaion of other
		 * modules if used.
		 *
		 * The cleaner way to do this is through the httpd.conf file:
		 *
		 *	# Disable the previous CustomLog directive.
		 * 	#CustomLog logs/access_log common
		 *	SetEnvIf Request_URI "/watch-info$" IGNORE
		 *	CustomLog /dev/null common env=IGNORE
		 *	CustomLog logs/access_log common env=!IGNORE
		 */
		return (isWatchHandler & watch_not_logged) == 0 ? OK : r->status;
	}

	/* Do not count if the connection is excluded. */
	if (ntIsMember(networkExcludeList, &r->connection->remote_addr.sin_addr)
	&& !ntIsMember(networkIncludeList, &r->connection->remote_addr.sin_addr))
		return DECLINED;

	/* Find last sub-request for bytes actually sent. */
	for ( ; r->next != (request_rec *) 0; r = r->next)
		;

	watchCounters(r);

	if (watch_what & WATCH_OWNER)
		(*watchUpdate)(r, watchGetUidName(r->pool, r->finfo.st_uid));

	if (watch_what & WATCH_VHOST)
		(*watchUpdate)(r, ap_get_server_name(r));

	if (watch_what & WATCH_SERVER)
		(*watchUpdate)(r, SERVER);

	if (watch_what & WATCH_REMOTE_IP)
		(*watchUpdate)(r, r->connection->remote_ip);

	if (dconf != (struct watchConfDir *) 0 && dconf->index != UNSET)
		(*watchUpdate)(r, ((char **) watchSpoolFileList->elts)[dconf->index]);

	return DECLINED;
}

/***********************************************************************
 ***  Log & Content using Weenie Files
 ***********************************************************************/

static FILE *
watchReadFile(request_rec *r, const char *file)
{
	int fd;
	FILE *fp;

	if ((file = ap_pstrcat(r->pool, spoolDirectoryRun, file, (char *) 0)) == (char *) 0)
		return (FILE *) 0;

	if ((fd = ap_popenf(r->pool, file, O_CREAT|O_RDWR, 0644)) < 0)
		return (FILE *) 0;

	if ((fp = ap_pfdopen(r->pool, fd, "r+")) == (FILE *) 0) {
		ap_pclosef(r->pool, fd);
		return (FILE *) 0;
	}

	ifInOctets = ifOutOctets = ifRequests = ifDocuments = 0;
	(void) fscanf(
		fp, "%lu %lu %lu %lu",
		&ifInOctets, &ifOutOctets, &ifRequests, &ifDocuments
	);

	return fp;
}

static void
watchUpdateFiles(request_rec *r, const char *file)
{
	FILE *fp;

	/* Use of flock() is a problem under SunOS and HP-UX.
	 * Use a real semaphore instead.
	 */
	if (MutexLock(mutex))
		goto error0;

	if ((fp = watchReadFile(r, file)) == (FILE *) 0)
		goto error1;

	rewind(fp);

	(void) fprintf(
		fp, "%lu %lu %lu %lu\n",
		ifInOctets += reqInOctets,
		ifOutOctets += reqOutOctets,
		ifRequests += reqRequests,
		ifDocuments += reqDocuments
	);

	ap_pfclose(r->pool, fp);
error1:
	(void) MutexUnlock(mutex);
error0:
}

static int
watchContentFiles(request_rec *r)
{
	int rc;
	FILE *fp;
	const char *file;
	const char *name = watchTarget(r);

	if ((rc = ap_discard_request_body(r)) != OK)
		return rc;

	ifInOctets = ifOutOctets = ifRequests = ifDocuments = 0;

	file = ap_pstrcat(r->pool, spoolDirectoryRun, name, (char *) 0);
	if ((fp = ap_pfopen(r->pool, file, "r")) != (FILE *) 0) {
		(void) fscanf(
			fp, "%lu %lu %lu %lu",
			&ifInOctets, &ifOutOctets, &ifRequests, &ifDocuments
		);
		ap_pfclose(r->pool, fp);
	}

	return watchContentPrint(r, name);
}

/***********************************************************************
 ***  Log & Content using Shared Memory Hash
 ***********************************************************************/

static void
watchUpdateHash(request_rec *r, const char *name)
{
	struct shEntry *entry = shGetLockedEntry(shtable, name);

	if (entry == (struct shEntry *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shGetLockedEntry(%lx, \"%s\") failed in watchUpdateHash()",
			(unsigned long) shtable, name
		);
		return;
	}

	entry->ifInOctets  += reqInOctets;
	entry->ifOutOctets += reqOutOctets;
	entry->ifRequests  += reqRequests;
	entry->ifDocuments += reqDocuments;

	if (shUnlock(shtable)) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shUnlock(%lx) failed in watchUpdateHash()",
			(unsigned long) shtable
		);
	}
}

static int
watchContentHash(request_rec *r)
{
	int rc;
	struct shEntry *entry;
	const char *name = watchTarget(r);

	if ((rc = ap_discard_request_body(r)) != OK)
		return rc;

	entry = shGetLockedEntry(shtable, name);

	if (entry == (struct shEntry *) 0) {
		ap_log_error(
			APLOG_MARK, APLOG_CRIT, watchMainServer,
			"shGetLockedEntry(%lx, \"%s\") failed in watchContentHash()",
			(unsigned long) shtable, name
		);

		ifInOctets  =
		ifOutOctets =
		ifRequests  =
		ifDocuments = 0;
	} else {
		ifInOctets  = entry->ifInOctets;
		ifOutOctets = entry->ifOutOctets;
		ifRequests  = entry->ifRequests;
		ifDocuments = entry->ifDocuments;

		if (shUnlock(shtable)) {
			ap_log_error(
				APLOG_MARK, APLOG_CRIT, watchMainServer,
				"shUnlock(%lx) failed in watchContentHash()",
				(unsigned long) shtable
			);
		}
	}

	return watchContentPrint(r, name);
}

static int
watchContentTable(request_rec *r)
{
	int i;
	DIR *dir;
	FILE *fp;
	unsigned long hash;
	struct dirent *dirent;
	struct shEntry *array;

	ap_send_http_header(r);

	ap_log_error(
		APLOG_MARK, APLOG_NOERRNO|APLOG_DEBUG, watchMainServer,
		"watchContentTable(%lx) shtable=%lx shtable.mutex=%lx mutex=%lx", r,
		(unsigned long) shtable, (unsigned long) shtable->mutex, (unsigned long) mutex
	);

	if (r->header_only)
		goto error0;

	if (shLock(shtable) && MutexLock(mutex))
		goto error0;

	ap_rprintf(r, "
<html>
<style type='text/css'>
th {
	font-family: Verdana, Arial, Helvetica, sans-serif;
	font-size: 10pt;
	background-color: #0080D7;
	color: #ffffff;
}
.normal, p, ul, td {
	font-family: Verdana, Arial, Helvetica, sans-serif;
	font-size: 10pt;
	color: #000000;
}
</style>
<body>
"
);

	ap_rprintf(r, "<p align='center'>"MODULE"/"VERSION"</p>\n");

	if (shtable != (struct shTable *) 0) {
		ap_rprintf(r, "<table width='100%%' border='0' cellpadding='0' cellspacing='0'>\n");
		ap_rprintf(r, "<tr align='center'>\n");
		ap_rprintf(r, "<th>Table Size</th><th>Found</th><th>Probes</th><th>Faults</th><th>Flushes</th><th>Shared Memory Remaining</th>\n");
		ap_rprintf(r, "</tr>\n");
		ap_rprintf(r, "<tr align='center'>\n");
		ap_rprintf(
			r, "<td>%d</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td>\n",
			shtable->size, shtable->info->found, shtable->info->probes,
			shtable->info->faults, shtable->info->flushes,
			MemoryAvailable(shtable->shared)

		);
		ap_rprintf(r, "</tr>\n");
		ap_rprintf(r, "</table>\n");
	}

	ap_rprintf(r, "<table width='100%%' border='0' cellpadding='1' cellspacing='0'>\n");
	ap_rprintf(r, "<tr align='right'>\n");
	ap_rprintf(r, "<th>Key</th><th>Hash</th><th>Optimal</th><th>Index</th><th>OutOctets</th><th>Requests</th><th>Documents</th>\n");
	ap_rprintf(r, "</tr>\n");

	if (shtable != (struct shTable *) 0) {
		array = shtable->array;
		for (i = 0; i < shTableSize; ++i) {
			if (array[i].key == (char *) 0)
				continue;

			ap_rprintf(r, "<tr align='right'>\n");
			ap_rprintf(
				r, "<td>%s</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td><td>%lu</td>\n",
				array[i].key,
				hash = shHashCode(0, array[i].key),
				hash % shTableSize,
				i,
				array[i].ifOutOctets,
				array[i].ifRequests,
				array[i].ifDocuments
			);
			ap_rprintf(r, "</tr>\n");
		}
	}

	/* List all weenie files NOT loaded into the hash table. */
	if ((dir = ap_popendir(r->pool, spoolDirectoryRun)) == (DIR *) 0)
		goto error1;

	while ((dirent = readdir(dir)) != (struct dirent *) 0) {
		if (*dirent->d_name == '.' || strcmp(dirent->d_name, shLockFile) == 0)
			continue;

		if (shContainsKey(shtable, dirent->d_name))
			continue;

		if ((fp = watchReadFile(r, dirent->d_name)) == (FILE *) 0)
			continue;

		/* The variables ifInOctets, ifOutOctets, ifRequests, ifDocuments
		 * will have already been filled in from the weenie file.
		 */
		ap_pfclose(r->pool, fp);

		ap_rprintf(r, "<tr align='right'>\n");
		ap_rprintf(
			r, "<td>%s</td><td>%lu</td><td>%lu</td><td>--</td><td>%lu</td><td>%lu</td><td>%lu</td>\n",
			dirent->d_name,
			hash = shHashCode(0, dirent->d_name),
			hash % shTableSize,
			ifOutOctets,
			ifRequests,
			ifDocuments
		);
		ap_rprintf(r, "</tr>\n");
	}

	ap_pclosedir(r->pool, dir);
error1:
	ap_rprintf(r, "</table>\n");
	ap_rprintf(r, "</body></html>\n");
	shUnlock(shtable) && MutexUnlock(mutex);
error0:
	return OK;
}

/***********************************************************************
 ***  Configuration Directives
 ***********************************************************************/

/*
 * Allows the URL used to query virtual host data:
 *
 *	http://www.snert.com/watch-info
 *
 *	<Location /watch-info>
 *	SetHandler watch-info
 *	</Location>
 *
 * Allows the URL used to query file owner and web
 * server data:
 *
 *	http://www.snert.com/~achowe/watch-info
 *	http://www.snert.com/~SERVER/watch-info
 *
 *	<Location /~* /watch-info>
 *	SetHandler watch-info
 *	</Location>
 *
 *	<Location /~* /watch-info>
 *	SetHandler watch-info
 *	</Location>
 *
 * The following allows me to verify the version and copyright:
 *
 *	<Location /watch-copyright>
 *	SetHandler watch-copyright
 *	Order allow,deny
 *	Allow from snert.biz snert.com snert.info snert.net snert.org
 *	</Location>
 */
static handler_rec watchHandlers[] = {
	{ "watch-info", watchContentHash },
	{ "watch-table", watchContentTable },
	{ "watch-copyright", watchContentCopyright },
	{ NULL }
};

/*
 * WatchHashTableSize number
 *
 * The size of the shared memory hash table for the "hash" storage policy.
 */
static const char *
WatchHashTableSize(cmd_parms *cmd, void *dconf, char *number)
{
	return setUnsignedInt(number, &shTableSize);
}

/*
 * WatchSpoolDirectory directory
 *
 * Absolute or server root relative directory where support and runtime
 * files are kept.
 */
static const char *
WatchSpoolDirectory(cmd_parms *cmd, void *dconf, char *directory)
{
	return setDirectory(cmd->pool, directory, &spoolDirectoryConf);
}

/*
 * WatchStorage files|hash
 */
static const char *
WatchStorage(cmd_parms *cmd, void *dconf, const char *policy)
{
	long value;
	store_policy *dp;
	const char *msg;

	for (dp = policy_table; dp->name != (const char *) 0; ++dp) {
		if (ap_strcasecmp_match(policy, dp->name) == 0)
			break;
	}

	if (dp->name == (const char *) 0)
		return "Invalid storage policy.";

	/* Set the handler to be used for logging. */
	watchUpdate = dp->update;
	watchHandlers[0].handler = dp->content;

	return (const char *) 0;
}

/*
 * WatchDocuments mime-type ...
 */
static const char *
WatchDocuments(cmd_parms *cmd, void *dconf, char *arg)
{
	char **mime;

	if (document_conf == (array_header *) 0)
		document_conf = ap_make_array(cmd->pool, 0, sizeof *documents_def);

	mime = (char **) ap_push_array(document_conf);
	if (mime == (char **) 0)
		return "Cannot add to WatchDocuments list.";

	*mime = (char *) ap_pstrdup(cmd->pool, arg);

	return (const char *) 0;
}

/*
 * WatchSpoolFile unique-filename
 */
static const char *
WatchSpoolFile(cmd_parms *cmd, void *dconf, char *filename)
{
	int i;
	char **entry, **table;

	if (watchSpoolFileConf == (array_header *) 0)
		watchSpoolFileConf = ap_make_array(cmd->pool, 0, sizeof (char *));

	table = (char **) watchSpoolFileConf->elts;
	for (i = 0; i < watchSpoolFileConf->nelts; ++i) {
		if (strcmp(filename, table[i]) == 0)
			return "WatchSpoolFile argument is not unique.";
	}

	((struct watchConfDir *) dconf)->index = watchSpoolFileConf->nelts;

	entry = (char **) ap_push_array(watchSpoolFileConf);
	if (entry == (char **) 0)
		return "Cannot add to WatchSpoolFile list.";

	*entry = (char *) ap_pstrdup(cmd->pool, filename);

	return (const char *) 0;
}

/*
 * WatchWhat item ...
 */
static const char *
WatchWhat(cmd_parms *cmd, void *dconf, char *arg)
{
	/* Erase the default settings when explicitly set. */
	if (watch_what & WATCH_DEFAULT)
		watch_what = 0;

	if (ap_strcasecmp_match(arg, "nothing") == 0)
		watch_what = 0;
	else if (ap_strcasecmp_match(arg, "virtual-host") == 0)
		watch_what |= WATCH_VHOST;
	else if (ap_strcasecmp_match(arg, "file-owner") == 0)
		watch_what |= WATCH_OWNER;
	else if (ap_strcasecmp_match(arg, "server") == 0)
		watch_what |= WATCH_SERVER;
	else if (ap_strcasecmp_match(arg, "remote-ip") == 0)
		watch_what |= WATCH_REMOTE_IP;
	else if (ap_strcasecmp_match(arg, "all") == 0)
		watch_what |= WATCH_VHOST|WATCH_OWNER|WATCH_SERVER|WATCH_REMOTE_IP;
	else
		return "Invalid item for WatchWhat";

	return (const char *) 0;
}

/*
 * WatchNotLogged handler ...
 */
static const char *
WatchNotLogged(cmd_parms *cmd, void *dconf, char *arg)
{
	/* Erase the default settings when explicitly set. */
	if (watch_not_logged == ~0)
		watch_not_logged = 0;

	if (ap_strcasecmp_match(arg, "watch-copyright") == 0)
		watch_not_logged |= WATCH_COPYRIGHT;
	else if (ap_strcasecmp_match(arg, "watch-info") == 0)
		watch_not_logged |= WATCH_INFO;
	else if (ap_strcasecmp_match(arg, "watch-table") == 0)
		watch_not_logged |= WATCH_TABLE;
	else if (ap_strcasecmp_match(arg, "all") == 0)
		watch_not_logged |= WATCH_COPYRIGHT|WATCH_INFO|WATCH_TABLE;
	else
		return "Invalid handler name for WatchNotLogged";

	return (const char *) 0;
}

/*
 * WatchNetwork {include|exclude} ip/mask ...
 */
static const char *
WatchNetwork(cmd_parms *cmd, void *dconf, char *set, char *network)
{
	if (networkIncludeConf == (array_header *) 0)
		networkIncludeConf = ap_make_array(cmd->pool, 0, sizeof (struct network));

	if (networkExcludeConf == (array_header *) 0)
		networkExcludeConf = ap_make_array(cmd->pool, 0, sizeof (struct network));

	if (ap_strcasecmp_match(set, "include") == 0)
		return ntAddNetwork(networkIncludeConf, network);
	else if (ap_strcasecmp_match(set, "exclude") == 0)
		return ntAddNetwork(networkExcludeConf, network);

	return "Invalid category for WatchNetwork";
}

static command_rec watchCommands[] = {
	{ "WatchDocuments", WatchDocuments, NULL, RSRC_CONF, ITERATE,
	"List of MIME types considered to be documents." },

	{ "WatchSpoolFile", WatchSpoolFile, NULL, ACCESS_CONF, TAKE1,
	"A unique filename per <Directory> or <Location> that is to be watched." },

	{ "WatchHashTableSize", WatchHashTableSize, NULL, RSRC_CONF, TAKE1,
	"Shared memory hash table size for \"hash\" storage policy." },

	{ "WatchNotLogged", WatchNotLogged, NULL, RSRC_CONF, ITERATE,
	"Do not log one or more handlers: all, watch-info, watch-copyright, watch-table" },

	{ "WatchNetwork", WatchNetwork, NULL, RSRC_CONF, ITERATE2,
	"Include or exclude networks (IP/mask) from being watched." },

	{ "WatchSpoolDirectory", WatchSpoolDirectory, NULL, RSRC_CONF, TAKE1,
	"Spool directory for any support and runtime files." },

	{ "WatchStorage", WatchStorage, NULL, RSRC_CONF, TAKE1,
	"The storage policy: files, hash" },

	{ "WatchWhat", WatchWhat, NULL, RSRC_CONF, ITERATE,
	"One or more items to watch: all, virtual-host, file-owner, remote-ip, server." },

	{ NULL }
};

/***********************************************************************
 ***  Module Definition Block
 ***********************************************************************/

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT watch_module = {
	STANDARD_MODULE_STUFF,
	watchInit,		/* module initializer                  */
	watchCreateDir,		/* create per-dir    conf structures */
	NULL,			/* merge  per-dir    conf structures */
	NULL,			/* create per-server conf structures */
	NULL,			/* merge  per-server conf structures */
	watchCommands,		/* table of conf file commands       */
	watchHandlers,		/* [#8] MIME-typed-dispatched handlers */
	NULL,			/* [#1] URI to filename translation    */
	NULL,			/* [#4] validate user id from request  */
	NULL,			/* [#5] check if the user is ok _here_ */
	NULL,			/* [#3] check access by host address   */
	NULL,			/* [#6] determine MIME type            */
	watchFixup,		/* [#7] pre-run fixups                 */
	watchLog,		/* [#9] log a transaction              */
#if MODULE_MAGIC_NUMBER >= 19970103
	NULL,			/* [#2] header parser                  */
#endif
#if MODULE_MAGIC_NUMBER >= 19970719
	NULL,			/* child_init			       */
#endif
#if MODULE_MAGIC_NUMBER >= 19970728
	NULL,			/* child_exit			       */
#endif
#if MODULE_MAGIC_NUMBER >= 19970902
	NULL			/* [#0] post read-request              */
#endif
};

